import 'dart:async';
import 'dart:math';
import 'dart:typed_data';

import 'package:crypto/crypto.dart';
import 'package:cryptography/cryptography.dart' as cryptography;
import 'package:ed25519_edwards/ed25519_edwards.dart' as ed25519;
import 'package:x25519/x25519.dart' as x25519;
import 'codec.dart';

class Crypto {
  static Future<List<int>> encrypt(List<int> key, List<int> blob) async {
    var box = await cryptography.Xchacha20.poly1305Aead()
        .encrypt(blob, secretKey: cryptography.SecretKey(key));
    var text = box.cipherText;
    var mac = box.mac.bytes;
    var nonce = box.nonce;

    return Codec.encode([nonce, mac, text]);
  }

  static Future<List<int>> deriveKey(
      List<int> mySK, List<int> otherSidePK) async {
    return x25519.X25519(
        mySK.sublist(0, 32).toList(), otherSidePK.sublist(32).toList());
  }

  static Future<List<int>> randKey() async {
    return randBytes(32);
  }

  static Future<List<int>> decrypt(List<int> key, List<int> box) {
    var parts = Codec.decode(box);
    if (parts.length != 3) {
      throw Exception("ExInvalidCipherText");
    }
    var nonce = parts[0];
    var mac = parts[1];
    var text = parts[2];
    return cryptography.Xchacha20.poly1305Aead().decrypt(
        cryptography.SecretBox(text.toList(),
            nonce: nonce.toList(), mac: cryptography.Mac(mac.toList())),
        secretKey: cryptography.SecretKey(key));
  }

  static Future<List<int>> digest(List<int> blob) async {
    return sha256.convert(blob).bytes;
  }

  static Future<Uint8List> randBytes(int n) async {
    var src = Random.secure();
    return Uint8List.fromList(
        List<int>.generate(n, (index) => src.nextInt(256)));
  }

  static Future<List<int>> genSK() async {
    var pair = x25519.generateKeyPair();
    var sk = ed25519.newKeyFromSeed(Uint8List.fromList(pair.privateKey));
    var list = List<int>.filled(32 + 32 + 32, 0);
    list.setRange(0, 64, sk.bytes);
    list.setRange(64, 64 + 32, pair.publicKey);
    return list;
  }

  static List<int> derivePK(List<int> secretKey) {
    var sk = secretKey;
    if (sk.length < ed25519.PrivateKeySize) {
      throw ArgumentError("invalid secret key");
    }
    return sk.sublist(32);
  }

  static Future<List<int>> sign(List<int> sk, List<int> data) async {
    var key = ed25519.PrivateKey(sk.sublist(0, 64).toList());
    var sig = ed25519.sign(key, Uint8List.fromList(data));
    return sig;
  }

  static Future<bool> verify(
      Uint8List data, List<int> signature, List<int> pk) async {
    return ed25519.verify(ed25519.PublicKey(pk.sublist(0, 32)), data,
        Uint8List.fromList(signature));
  }
}
