import 'dart:async';
import 'package:async/async.dart';

import 'mem_cache.dart';

import 'crypto.dart';
import 'packet.dart';
import 'net_addr.dart';
import 'peer_info.dart';
import 'tail_io.dart';
import 'utils.dart';
import 'verival.dart';
import 'io.dart';
import 'cache.dart';

class Node {
  final List<int> sk;
  final List<Tail> tails;
  final Cache cache;
  bool _closed = false;

  Node(this.sk, this.tails, this.cache) {
    listen();
  }

  static Future<Node> create(List<int> sk, NetAddr tail, [Cache? cache]) async {
    return Node(
        sk, [await Tail.connect(tail)], cache is Cache ? cache : MemCache());
  }

  void close() {
    if (_closed) {
      return;
    }
    for (var tail in tails) {
      try {
        tail.close();
      } catch (_) {}
    }
    _closed = true;
  }

  bool get closed {
    return _closed;
  }

  Future<bool> _setLastPacket(Packet packet, String instance) async {
    var last = await _getLastPacket(instance);
    if (last is Packet && last.id > packet.id) {
      return false;
    }
    await cache.set("last_packet_$instance", packet.encode());
    return true;
  }

  Future<Packet?> _getLastPacket(String instance) async {
    try {
      return Packet.decode((await cache.get("last_packet_$instance"))!);
    } catch (error) {
      return null;
    }
  }

  Stream<Input> get stream {
    return StreamGroup.merge(
            tails.map((tail) => tail.stream.asyncMap((dg) async {
                  Packet packet;
                  try {
                    packet = Packet.decode(dg);
                  } catch (_) {
                    return null;
                  }
                  if (!Utils.equal(packet.to, pk())) {
                    return null;
                  }
                  await _setLastPacket(packet, tail.host.toString());
                  return packet;
                })))
        .where((packet) => packet is Packet)
        .asyncMap((packet) async {
          try {
            return await Input.decrypt(packet!, sk);
          } catch (_) {
            return null;
          }
        })
        .where((msg) => msg is Input)
        .map((msg) => msg!);
  }

  Duration delay = Duration(milliseconds: 50);

  List<int> pk() {
    return Crypto.derivePK(sk);
  }

  void listen() {
    for (var tail in tails) {
      Timer.run(() async {
        while (!_closed) {
          try {
            await Future.delayed(delay);
            var last = (await _getLastPacket(tail.host.toString())) ??
                Packet(pk(), List.filled(pk().length, 0), [], 1);
            await tail.check(last.encode());
          } catch (_) {}
        }
      });
    }
  }

  Future<void> publish(String name) {
    return Verival.publish(PeerInfo(tail: tails.first.host, name: name), sk);
  }

  Future<void> send(Output out) async {
    var peerInfoFuture = Verival.lookup(out.to, Duration(seconds: 5));
    var packet = await out.encrypt(sk, -1);
    var encoded = packet.encode();
    if (encoded.length > 400) {
      peerInfoFuture.ignore();
      throw Exception("too large packet");
    }
    var peerInfo = await peerInfoFuture;
    await Tail.push(peerInfo.tail, encoded, sk, Duration(seconds: 5));
  }
}
